package org.processmining.localprocessmodelconformance.plugins;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeDiscreteImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;

@Plugin(name = "Run CloGSgrow", parameterLabels = { "Log" }, returnLabels = { "CloGSgrow patterns" }, returnTypes = {
		XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class ClosGrow {
	private int min_sup;
	private XLog Fre;
	private Set<String> mathcalE;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Run CloGSgrow", requiredParameterLabels = { 0 })
	public XLog abstractLogBasedOnPatterns(PluginContext context, XLog log){
		return runHeadless(log, 5);
	}
	
	public XLog runHeadless(XLog log, int min_sup){
		this.min_sup = min_sup;
		// Algorithm 3, line 1
		this.Fre = newEmptyLog();
		this.mathcalE = new HashSet<String>();
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, new XEventNameClassifier());
		for(XEventClass xec : logInfo.getEventClasses().getClasses())
			mathcalE.add(xec.getId());
		
		// Algorithm 3, line 2
		for(String e : mathcalE){
			// Algorithm 3, line 3
			XTrace P = newEmptyTrace();
			P.add(newEventWithActivity(e));
			XLog I = newEmptyLog();
			for(int i=0; i<log.size(); i++){
				XTrace t = log.get(i);
				for(int j=0; j<t.size(); j++){
					XEvent eIns = t.get(j);
					if(getActivity(eIns).equals(e)){
						XTrace ti = newTraceWithIndex(i);
						XEvent l = newEventWithActivityAndIndex(e, j);
						ti.add(l);
						I.add(ti);
					}
				}
			}
			// Algorithm 3, line 4
			mineFre(log, P, I);
		}
		// Algorithm 3, line 5
		return Fre;
	}
	
	private void mineFre(XLog log, XTrace P, XLog I) {
		// Algorithm 3, line 6
		if(I.size()>=min_sup){
			P.getAttributes().put("support", new XAttributeDiscreteImpl("support", I.size()));
			P.getAttributes().put("instances", new XAttributeLiteralImpl("instances", getInstancesAsString(I)));
			if(CCheck(P)){
			// Algorithm 3, line 7
				Fre.add(P);
			}
			// Algorithm 3, line 8
			for(String e : mathcalE){
				// Algorithm 3, line 9
				XLog Iplus = INSgrow(log, P, I, newEventWithActivity(e));
				// Algorithm 3, line 10
				XTrace PNew = (XTrace) P.clone();
				PNew.add(newEventWithActivity(e));
				mineFre(log, PNew, Iplus);
			}
		}
	}

	// true represents "closed", false represents "non-closed"
	private boolean CCheck(XTrace p) {
		Set<XTrace> nonClosed = new HashSet<XTrace>();
		boolean isNonClosed = false;
		for(XTrace fre : Fre){
			if(getSupport(p) >= getSupport(fre) && isExtension(p, fre)){
				nonClosed.add(fre);
			}else{
				if(getSupport(fre) >= getSupport(p) && isExtension(fre, p)){
					isNonClosed = true;
				}
			}
		}
		Fre.removeAll(nonClosed);
		return !isNonClosed;
	}

	private int getSupport(XTrace p) {
		return (int) ((XAttributeDiscrete) p.getAttributes().get("support")).getValue();
	}

	private boolean isExtension(XTrace t1, XTrace t2) {
		if((t1.size()-1)==t2.size()){
			int offset = 0;
			for(int i=0; i<t2.size(); i++){
				boolean matched = false;
				while(!matched){
					matched = getActivity(t1.get(i+offset)).equals(getActivity(t2.get(i)));
					if(!matched){
						if(offset==0){
							offset++;
						}else{
							return false;
						}
					}
				}
			}
			return true;
		}else{
			return false;
		}
	}

	private String getInstancesAsString(XLog I) {
		StringBuilder sb = new StringBuilder();
		for(XTrace i : I){
			sb.append('('); 
			sb.append(i.getAttributes().get("index".toString()));
			sb.append(", ");
			sb.append('[');
			boolean first = true;
			for(XEvent e : i){
				if(!first)
					sb.append(',');
				sb.append(e.getAttributes().get("index").toString());
				first = false;
			}
			sb.append("]) ");
		}
		return sb.toString();
	}

	public static XLog supComp(XLog log, XTrace pattern){
		// Algorithm 1, line 1
		XLog I = newEmptyLog();
		for(int i=0; i<log.size(); i++){
			XTrace trace = log.get(i);
			for(int j=0; j<trace.size(); j++){
				String activity = getActivity(pattern, 0);
				if(getActivity(trace, j).equals(activity)){
					XTrace t = newTraceWithIndex(i);
					t.add(newEventWithActivityAndIndex(activity, j));
					I.add(t);
				}
			}
		}
		// Algorthim 1, lines 2 & 3
		for(int j=1; j<pattern.size()-1; j++){
			XTrace subPattern = (XTrace) pattern.clone();
			for(int i=j+1; i<pattern.size(); i++){
			//for(int i = pattern.size(); i>=j; i++){
				subPattern.remove(j+1);
			}
			I = INSgrow(log, subPattern, I, pattern.get(j+1));
		}
		// Algorithm 1, line 4
		return I;
	}
	
	public static XLog INSgrow(XLog log, XTrace subPattern, XLog I, XEvent e){
		// Algorithm 2, line 1
		XLog IiPlusUnion = newEmptyLog();
		for(int i=0; i<log.size(); i++){
			XTrace Si = log.get(i);
			if(!hasInstance(subPattern, Si))
				continue;
			XLog Ii = newEmptyLog();
			for(XTrace iTrace : I){
				if(i==getIndexOfTrace(iTrace)){
					Ii.add(iTrace);
				}
			}
			if(Ii==null)
				continue;
			// Algorithm 2, line 2
			int last_position = 0;
			// Algorthim 2, line 3
			Collections.sort(Ii, new RightShiftOrder());
			for(XTrace t : Ii){
				// Algorithm 2, line 4
				int lj = next(Si, e, Math.max(last_position, getIndexOfEvent(t.get(t.size()-1))));
				// Algorithm 2, line 5
				if(lj==Integer.MAX_VALUE)
					break;
				// Algorithm 2, line 6
				last_position = lj;
				// Algorithm 2, line 7
				XTrace tNew = (XTrace) t.clone();
				tNew.add(newEventWithActivityAndIndex(getActivity(e), lj));
				IiPlusUnion.add(tNew);
			}
		}
		return IiPlusUnion;	
	}
	
	private static boolean hasInstance(XTrace subPattern, XTrace si) {
		int traceIndex = 0;
		boolean found = true;
		for(XEvent e : subPattern){
			String matchActivity = getActivity(e);
			boolean localFound = false;
			while(!localFound){
				if(getActivity(si.get(traceIndex)).equals(matchActivity)){
					localFound = true;
					break;
				}
				if(traceIndex>=si.size()-1 && !localFound){
					found = false;
					break;
				}
				traceIndex++;
			}
			if(!found)
				break;
		}
		return found;
	}

	public static int next(XTrace Si, XEvent e, int lowest){
		int l=lowest+1;
		while(l<Si.size()){
			if(getActivity(Si.get(l)).equals(getActivity(e)))
				break;
			l++;
		}
		if(l==Si.size())
			l = Integer.MAX_VALUE;
		return l;
	}
	
	public static class RightShiftOrder implements Comparator<XTrace> {
		public int compare(XTrace t1, XTrace t2) {
			int it1 = getIndexOfTrace(t1);
			int it2 = getIndexOfTrace(t2);
			if(it1!=it2)
				return Integer.compare(it1, it2);
			int hT1 = getIndexOfEvent(t1.get(t1.size()-1));
			int hT2 = getIndexOfEvent(t2.get(t2.size()-1));
			return Integer.compare(hT1, hT2);
		}
	}
	
	public static int getIndexOfTrace(XTrace trace){
		return (int) ((XAttributeDiscrete) trace.getAttributes().get("index")).getValue();
	}
	
	public static int getIndexOfEvent(XEvent event){
		return (int) ((XAttributeDiscrete) event.getAttributes().get("index")).getValue();
	}
	
	public static XEvent newEventWithActivityAndIndex(String activity, int index){
		XEvent newEvent = newEventWithActivity(activity);
		newEvent.getAttributes().put("index", new XAttributeDiscreteImpl("index", index));
		return newEvent;
	}
	
	public static XEvent newEventWithActivity(String activity){
		XEvent newEvent = XFactoryRegistry.instance().currentDefault().createEvent();
		newEvent.getAttributes().put("concept:name", new XAttributeLiteralImpl("concept:name", activity));
		return newEvent;
	}
	
	public static XEvent newEmptyEvent(){
		return XFactoryRegistry.instance().currentDefault().createEvent();
	}
	
	public static XLog newEmptyLog(){
		return XFactoryRegistry.instance().currentDefault().createLog();
	}
	
	public static XTrace newTraceWithIndex(int index){
		XTrace newTrace = newEmptyTrace();
		newTrace.getAttributes().put("index", new XAttributeDiscreteImpl("index", index));
		return newTrace;
	}
	
	public static XTrace newEmptyTrace(){
		return XFactoryRegistry.instance().currentDefault().createTrace();
	}
	
	public static String getActivity(XTrace trace, int position){
		return getActivity(trace.get(position));
	}
	
	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
}
