package org.processmining.localprocessmodelconformance.plugins;

import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.logenhancement.abstraction.PatternStructureException;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;

@Plugin(name = "Select Local Process Models Greedily", parameterLabels = { "Log",
"Local Process Models" }, returnLabels = { "Local Process Models" }, returnTypes = {
		LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class GreedyLPMSelector {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Select Local Process Models Greedily", requiredParameterLabels = { 0, 1 })
	public LocalProcessModelRanking greedySelector(PluginContext context, XLog log, LocalProcessModelRanking ranking) throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException, CompositionVisitorException {
		Petrinet[] nonAcceptingNets = new Petrinet[ranking.getSize()];
		for(int i=0; i<ranking.getSize(); i++){
			nonAcceptingNets[i] = ranking.getNet(i).getAcceptingPetriNet().getNet();
		}
		return greedySelector(log, nonAcceptingNets, new HashSet<Petrinet>(), 5);
	}

	private static LocalProcessModelRanking greedySelector(XLog log, Petrinet[] nets, Set<Petrinet> selectedNets, int minimumMatchedEventThreshold) throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException, CompositionVisitorException {
		int maxMatchedEventCount = 0;
		Petrinet candidateNet = null;
		XLog newLog = null;
		Set<Petrinet> netsThatFullfillTheThreshold = new HashSet<Petrinet>();
		for(Petrinet net : nets){
			Petrinet[] singleton = new Petrinet[1];
			singleton[0] = net;
			Object[] result = RetrieveLPMEpisodesForLPMCollection.matchPatternsToLogEpisodes(log, net);
			int matchedEventCount = (int) result[3];
			if(matchedEventCount>maxMatchedEventCount){
				maxMatchedEventCount = matchedEventCount;
				candidateNet = net;
				newLog = (XLog) result[0];
			}
			if(matchedEventCount>minimumMatchedEventThreshold)
				netsThatFullfillTheThreshold.add(net);
		}
		if(maxMatchedEventCount>=minimumMatchedEventThreshold){
			Petrinet[] newNets = new Petrinet[netsThatFullfillTheThreshold.size()-1];
			int i = 0;
			for(Petrinet net : netsThatFullfillTheThreshold){
				if(net!=candidateNet){
					newNets[i] = net;
					i++;
				}
			}
			selectedNets.add(candidateNet);
			return greedySelector(newLog, newNets, selectedNets, minimumMatchedEventThreshold);
		}else{
			LocalProcessModelRanking result = new LocalProcessModelRanking();
			for(Petrinet net : selectedNets)
				result.addElement(new LocalProcessModel(new AcceptingPetriNetImpl(net)));
			return result;
		}
	}

}
