package org.processmining.localprocessmodelconformance.plugins;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.lpm.util.LocalProcessModel;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;

@Plugin(name = "Make Local Process Model Ranking Closed", parameterLabels = { "Local Process Model Ranking" }, returnLabels = { "Closed Local Process Model Ranking" }, returnTypes = {LocalProcessModelRanking.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class MakeLocalProcessModelRankingClosed {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Make Local Process Model Ranking Closed", requiredParameterLabels = { 0 })
	public LocalProcessModelRanking greedySelector(PluginContext context, LocalProcessModelRanking ranking) {
		LocalProcessModelRanking closedRanking = new LocalProcessModelRanking();
		for(int i=0; i<ranking.getSize(); i++){
			boolean existsDominator = false;
			for(int j=0; j<ranking.getSize(); j++){
				if(i==j)
					continue;
				if(dominates(ranking.getNet(j), ranking.getNet(i))){
					existsDominator = true;
					break;
				}
			}
			if(!existsDominator)
				closedRanking.addElement(ranking.getNet(i));
		}
		System.out.println("Closed ranking size: "+closedRanking.getSize());
		return closedRanking;
	}

	/*
	 * One LPM (net) dominates another LPM (other) if it has equal or higher support, and if it is behaviorally more precise, meaning that either:
	 * 	1) the language of net is a subset of the language of other
	 *  2) all language elements (called the extensions) of net are a sequential extension of language elements (called the originators) of other
	 */
	private boolean dominates(LocalProcessModel net, LocalProcessModel other) {
		/* TODO: a better definition of closed based on process trees, using the language properties of the operators
		if(net.getProcessTree()!=null && other.getProcessTree()!=null){
			return treeDominates(net.getProcessTree(), other.getProcessTree());
		}else{ // fall back on language-based detection
		*/
			if(net.getFrequency()<other.getFrequency())
				return false;
			// test condition (1)
			if(other.getLanguage().containsAll(net.getLanguage()) && net.getLanguage().size()<other.getLanguage().size())
				return true;
			// test condition (2)
			for(List<String> languageElement : net.getLanguage()){
				boolean originatorFound = false;
				for(List<String> languageElement2 : other.getLanguage()){
					if(Collections.indexOfSubList(languageElement2, languageElement)!=-1){
						originatorFound = true;
						break;
					}
				}
				if(!originatorFound)
					return false;
			}
			return true;
		//}
	}

	private boolean treeDominates(ProcessTree netTree, ProcessTree otherTree) {
		// TODO: a better definition of closed based on process trees, using the language properties of the operators
		Set<Pair<Node, Node>> misAlignedNodePairs = new HashSet<Pair<Node, Node>>();
		Boolean validExtension = true;
		traverse(netTree.getRoot(), otherTree.getRoot(), misAlignedNodePairs);
		return false;
	}

	private boolean traverse(Node netNode, Node otherNode, Set<Pair<Node, Node>> misAlignedNodePairs) {
		/*
		 * TODO: a better definition of closed based on process trees, using the language properties of the operators
		if(netNode instanceof ManualTask && otherNode instanceof ManualTask){
			if(!((ManualTask) netNode).getName().equals(((ManualTask) otherNode).getName())){
				misAlignedNodePairs.add(new Pair<Node, Node>(netNode, otherNode));
			}
		}
		if(netNode instanceof AbstractBlock && otherNode instanceof AbstractBlock){
			if((netNode instanceof AbstractBlock.And && otherNode instanceof AbstractBlock.And) ||
				(netNode instanceof AbstractBlock.Xor && otherNode instanceof AbstractBlock.Xor) ||
				(netNode instanceof AbstractBlock.XorLoop && otherNode instanceof AbstractBlock.XorLoop) ||
				(netNode instanceof AbstractBlock.Seq && otherNode instanceof AbstractBlock.Seq) ||
				(netNode instanceof AbstractBlock.Or && otherNode instanceof AbstractBlock.Or)){
				List<Node> netChildren = new ArrayList<Node>();
				List<Node> otherChildren = new ArrayList<Node>();
				
				for(Node netNode2 : netNode.getProcessTree().getNodes()){
					if(netNode2.getParents().contains(netNode))
						netChildren.add(netNode2);
				}
				if(netNode.getIncomingEdges().size()!=otherNode.getIncomingEdges().size())
					return false;
				for(int i=0; i<netEdges.size(); i++){
					if(!traverse(netEdges.get(i).getTarget(), otherEdges.get(i).getTarget(), misAlignedNodePairs))
						return false;
				}
			}else{// nodes are NOT of same operator type
				return false;
			}
		}
		*/
		return false;
	}

}
