package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.plugins.utils.ProvidedObjectHelper;

@Plugin(name = "Make Log Per Non-Dominated Alphabet", parameterLabels = { "Log" }, returnLabels = { "Log for nonempty result - DO NOT USE" }, returnTypes = { XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement })
public class MakeLogPerNonDominatedAlphabet {
	public static XLog newEmptyLog(){
		return XFactoryRegistry.instance().currentDefault().createLog();
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Make Log Per Non-Dominated Alphabets", requiredParameterLabels = { 0 })
	public XLog run(PluginContext context, XLog log) {
		for(Pair<Set<String>, XLog> entry : runHeadless(log)){
			ProvidedObjectHelper.publish(context, "Log for alphabet: "+entry.getFirst(), entry.getSecond(), XLog.class, false);
		}
		return log;
	}

	private List<Pair<Set<String>, XLog>> runHeadless(XLog log) {
		Set<Set<String>> nonDominatedAlphabets = new HashSet<Set<String>>();
		for(XTrace t : log){
			Set<String> alphabet = getTraceAlphabet(t);
			if(!existingDominatingAlphabet(nonDominatedAlphabets, alphabet)){
				removeDominatedAlphabets(nonDominatedAlphabets, alphabet);
				nonDominatedAlphabets.add(alphabet);
			}
		}
		List<Pair<Set<String>, XLog>> resultList = new ArrayList<Pair<Set<String>, XLog>>();
		for(Set<String> nonDominatedAlphabet : nonDominatedAlphabets){
			XLog alphabetLog = newEmptyLog();
			for(XTrace t :log){
				Set<String> alphabet = getTraceAlphabet(t);
				if(nonDominatedAlphabet.containsAll(alphabet))
					alphabetLog.add(t);
			}
			resultList.add(new Pair<Set<String>, XLog>(nonDominatedAlphabet, alphabetLog));
		}
		return resultList;
	}

	private Set<String> getTraceAlphabet(XTrace t) {
		Set<String> alphabet = new HashSet<String>();
		for(XEvent e : t){
			alphabet.add(e.getAttributes().get("concept:name").toString());
		}
		return alphabet;
	}

	private boolean existingDominatingAlphabet(Set<Set<String>> nonDominatedAlphabets, Set<String> alphabet) {
		for(Set<String> existingAlphabet : nonDominatedAlphabets){
			if(existingAlphabet.containsAll(alphabet))
				return true;
		}
		return false;
	}
	
	private Set<Set<String>> removeDominatedAlphabets(Set<Set<String>> existingAlphabets, Set<String> alphabet) {
		Set<Set<String>> dominatedSet = new HashSet<Set<String>>();
		for(Set<String> existingAlphabet : existingAlphabets){
			if(alphabet.containsAll(existingAlphabet))
				dominatedSet.add(existingAlphabet);
		}
		existingAlphabets.removeAll(dominatedSet);
		return existingAlphabets;
	}

}
