package org.processmining.localprocessmodelconformance.plugins;

import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetArrayImpl;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.localprocessmodelconformance.plugins.LogAbstracterBasedOnAlignment.Pattern;
import org.processmining.logenhancement.abstraction.PatternStructureException;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.utils.ProvidedObjectHelper;

@Plugin(name = "Mine Split Miner Models for Log Episodes", parameterLabels = { "Log",
"Abstraction Patterns" }, returnLabels = { "Split Miner Petri Net Array" }, returnTypes = {AcceptingPetriNetArray.class }, userAccessible = true, categories = { PluginCategory.Discovery })
public class MineSplitMinerModelsForLogEpisodes {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Mine Split Miner Models for Log Episodes", requiredParameterLabels = { 0,
			1 })
	public AcceptingPetriNetArray abstractLogBasedOnPatterns(UIPluginContext context, XLog log, LocalProcessModelRanking ranking)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		Petrinet[] nonAcceptingNets = new Petrinet[ranking.getSize()];
		for(int i=0; i<ranking.getSize(); i++){
			nonAcceptingNets[i] = ranking.getNet(i).getAcceptingPetriNet().getNet();
		}
		Object[] results = RetrieveLPMEpisodesForLPMCollection.matchPatternsToLogEpisodes(context, log, nonAcceptingNets);
		Map<Pattern,XLog> episodeLogs = (Map<Pattern, XLog>) results[1];
		
		AcceptingPetriNetArray apa = new AcceptingPetriNetArrayImpl();
		for(Entry<Pattern,XLog> entry : episodeLogs.entrySet()){
			XLog episodeLog = entry.getValue();
			ProvidedObjectHelper.publish(context, "episodeLog: "+entry.getKey().toString(), episodeLog, XLog.class, false);

			if(episodeLog.size()==0)
				continue;
			Petrinet pn = SplitMinerWrapper.minePetrinetWithSplitMiner(context, episodeLog);
			AcceptingPetriNet apn = new AcceptingPetriNetImpl(pn);
			apa.addNet(apn);
		}
		context.getProvidedObjectManager().createProvidedObject("LPMs that occur in the log", (AcceptingPetriNetArray) results[2], AcceptingPetriNetArray.class, context);

		return apa;
	}

}
