package org.processmining.localprocessmodelconformance.plugins;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.lang.model.SourceVersion;
import javax.swing.JLabel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNet.PetrinetWithMarkings;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.localprocessmodelconformance.plugins.LogAbstracterBasedOnAlignment.Pattern;
import org.processmining.logenhancement.abstraction.PatternStructureException;
import org.processmining.logenhancement.abstraction.editor.AbstractionModelBuilder;
import org.processmining.logenhancement.abstraction.model.AbstractionModel;
import org.processmining.logenhancement.abstraction.model.AbstractionPattern;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.lpm.util.DummyPluginContext;
import org.processmining.lpm.util.LocalProcessModelRanking;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.BalancedDataXAlignmentPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverImpl;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignedLog;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;

import com.google.common.base.Function;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

@Plugin(name = "Check Conformance of LPM Collection", parameterLabels = { "Log",
"Abstraction Patterns" }, returnLabels = { "Abstracted Log" }, returnTypes = {
		XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement }, keywords = {
				"abstraction", "supervised",
				"pattern" }, help = "<html>Abstracts events from the supplied event log using a supervised pattern-based approach. "
						+ "This plug-in implements the methodology presented in the BPM'16 paper 'From Low-Level Events to Activities - A Pattern-based Approach' (DOI: <a href=\"http://dx.doi.org/10.1007/978-3-319-45348-4_8\">10.1007/978-3-319-45348-4_8)</a>."
						+ "Activity patterns describe the behavior expected for an (abstract) activity and should be modeled as Data Petri Net (DPN) with a single source and a single sink place (SESE stucture).</html>")

public class RetrieveLPMEpisodesForLPMCollection {
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Check Conformance of LPM Collection", requiredParameterLabels = { 0,
			1 })
	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, LocalProcessModelRanking ranking)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		AcceptingPetriNet[] nonAcceptingNets = new AcceptingPetriNet[ranking.getSize()];
		for(int i=0; i<ranking.getSize(); i++){
			nonAcceptingNets[i] = ranking.getNet(i).getAcceptingPetriNet();
		}
		return abstractLogBasedOnPatterns(context, log, nonAcceptingNets);
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Check Conformance of LPM Collection", requiredParameterLabels = { 0,
			1 })
	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, AcceptingPetriNet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		Petrinet[] nonAcceptingNets = new Petrinet[abstractionNets.length];
		for(int i=0; i<abstractionNets.length; i++){
			Petrinet candidateNet = abstractionNets[i].getNet();
			if(findSource(candidateNet)==null){
				Place newInitial = candidateNet.addPlace("source 1");
				Place originalInitial = abstractionNets[i].getInitialMarking().toArray(new Place[abstractionNets[i].getInitialMarking().size()])[0];
				Transition artificialSilentConnector = candidateNet.addTransition("artificialSilentConnector");
				artificialSilentConnector.setInvisible(true);
				candidateNet.addArc(newInitial, artificialSilentConnector);
				candidateNet.addArc(artificialSilentConnector, originalInitial);
			}
			nonAcceptingNets[i] = abstractionNets[i].getNet();
		}
		return abstractLogBasedOnPatterns(context, log, nonAcceptingNets);
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Check Conformance of LPM Collection", requiredParameterLabels = { 0,
			1 })
	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, Petrinet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		Object[] result = matchPatternsToLogEpisodes(context, log, abstractionNets);
		ProvidedObjectHelper.publish(context, "LPMs that occur in the log", (AcceptingPetriNetArray) result[2], AcceptingPetriNetArray.class, false);
		Map<Pattern,XLog> episodeLogs = (Map<Pattern, XLog>) result[1];

		for(Entry<Pattern,XLog> entry : episodeLogs.entrySet()){
			XLog episodeLog = entry.getValue();
			ProvidedObjectHelper.publish(context, "episodeLog: "+entry.getKey().toString(), episodeLog, XLog.class, false);
		}
		
		ProvidedObjectHelper.publish(context, "Annotated Log", (XLog) result[6], XLog.class, false);

		return (XLog)result[0];
	}

	public static Object[] matchPatternsToLogEpisodes(XLog log, Petrinet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException,
			CompositionVisitorException {
		DataPetriNet[] transformedNets = transformNets(abstractionNets);

		int i=1;
		Map<DataPetriNet, String> patternToName = new HashMap<DataPetriNet, String>();
		for(DataPetriNet dpn : transformedNets){
			patternToName.put(dpn, "pattern"+i);
			i++;
		}

		Map<String, AbstractionPattern> abstractionPatterns = buildPatterns(patternToName);

		String initialModel = buildInitialModel(abstractionPatterns);
		AbstractionModelBuilder modelBuilder = new AbstractionModelBuilder(abstractionPatterns);
		modelBuilder.update(initialModel);
		StringBuilder compositionBuilder = new StringBuilder();
		List<String> keys = new ArrayList<String>(abstractionPatterns.keySet());
		if(keys.size()>1)
			compositionBuilder.append("(%[\"");
		else
			compositionBuilder.append("\"");
		for(int j=0; j<keys.size(); j++){
			compositionBuilder.append(keys.get(j));
			compositionBuilder.append("\"");
			compositionBuilder.append("*");
			if((j+1)<keys.size())
				compositionBuilder.append(",\"");
		}
		if(keys.size()>1)
			compositionBuilder.append("])*");
		System.err.println(compositionBuilder.toString());
		modelBuilder.update(compositionBuilder.toString());
		return abstractPatterns(log, modelBuilder.getAbstractionModel());
	}
	
	public static Object[] matchPatternsToLogEpisodes(UIPluginContext context, XLog log, Petrinet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException, ParseException,
			CompositionVisitorException {
		DataPetriNet[] transformedNets = transformNets(abstractionNets);
		for(DataPetriNet dpn : transformedNets)
			ProvidedObjectHelper.publish(context, "Transformed net for episode matching", dpn, DataPetriNet.class, false);

		int i=1;
		Map<DataPetriNet, String> patternToName = new HashMap<DataPetriNet, String>();
		for(DataPetriNet dpn : transformedNets){
			patternToName.put(dpn, "pattern"+i);
			i++;
		}

		Map<String, AbstractionPattern> abstractionPatterns = buildPatterns(patternToName);

		String initialModel = buildInitialModel(abstractionPatterns);
		AbstractionModelBuilder modelBuilder = new AbstractionModelBuilder(abstractionPatterns);
		modelBuilder.update(initialModel);
		StringBuilder compositionBuilder = new StringBuilder();
		List<String> keys = new ArrayList<String>(abstractionPatterns.keySet());
		if(keys.size()>1)
			compositionBuilder.append("(%[\"");
		else
			compositionBuilder.append("\"");
		for(int j=0; j<keys.size(); j++){
			compositionBuilder.append(keys.get(j));
			compositionBuilder.append("\"");
			compositionBuilder.append("*");
			if((j+1)<keys.size())
				compositionBuilder.append(",\"");
		}
		if(keys.size()>1)
			compositionBuilder.append("])*");
		System.err.println(compositionBuilder.toString());
		modelBuilder.update(compositionBuilder.toString());
		try {
			return abstractPatterns(context, log, modelBuilder.getAbstractionModel());
		} catch (UserCancelledException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public Object[] abstractLogBasedOnPatterns(UIPluginContext context, XLog log, PetrinetGraph... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {

		try {
			InteractionResult result = context
					.showConfiguration("Abstract Low-Level Events to Activities using Patterns",
							new JLabel(""));
			if (result == InteractionResult.CANCEL) {
				throw new UserCancelledException();
			}

			DataPetriNet[] transformedNets = transformNets(abstractionNets);
			for(DataPetriNet dpn : transformedNets)
				ProvidedObjectHelper.publish(context, "Transformed net for episode matching", dpn, DataPetriNet.class, false);
			Map<DataPetriNet, String> patternToName = ProMUIHelper.queryMapToString(context,
					"Configure the mapping between activity patterns (left) and the desired names for the abstracted events (right)",
					transformedNets,
					Maps.asMap(ImmutableSet.copyOf(transformedNets), new Function<DataPetriNet, String>() {

						public String apply(DataPetriNet net) {
							return net.getLabel().replaceAll(".pnml", "");
						}
					}));

			Map<String, AbstractionPattern> abstractionPatterns = buildPatterns(patternToName);

			String initialModel = buildInitialModel(abstractionPatterns);
			AbstractionModelBuilder modelBuilder = new AbstractionModelBuilder(abstractionPatterns);
			modelBuilder.update(initialModel);

			result = context.showConfiguration("Configure the abstraction model (restrict the possible interaction)",
					modelBuilder);
			while (!modelBuilder.isValid() && result != InteractionResult.CANCEL) {
				result = context.showConfiguration("Invalid Model: Specify a valid abstraction model!", modelBuilder);
			}
			if (result != InteractionResult.CANCEL) {
				return abstractPatterns(context, log, modelBuilder.getAbstractionModel());
			} else {
				context.getFutureResult(0).cancel(false);
				return null;
			}
		} catch (UserCancelledException e) {
			context.getFutureResult(0).cancel(false);
			return null;
		}
	}
	
	public static String buildInitialModel(Map<String, AbstractionPattern> abstractionPatterns) {
		StringBuilder sb = new StringBuilder();
		for (Iterator<String> iter = abstractionPatterns.keySet().iterator(); iter.hasNext();) {
			String id = iter.next();
			if (!SourceVersion.isIdentifier(id)) {
				id = "\"" + id + "\"";
			}
			sb.append(id);
			sb.append("*");
			if (iter.hasNext()) {
				sb.append('#');
			}
		}
		return sb.toString();
	}
	
	private static Place findSourceByStructure(PetrinetGraph net) {
		Place source = null;
		for (Place p : net.getPlaces()) {
			if (net.getInEdges(p).isEmpty()) {
				if (source == null) {
					source = p;
				} else {
					return null;
				}
			}
		}
		return source;
	}
	
	private static Transition findStartTransition(PetrinetGraph net) {
		Place source = findSourceByStructure(net);
		if (source == null){
			Place artificialSource = net.addPlace("artificialSource");
			Transition silentConnector = net.addTransition("artificialSilentConnector");
			silentConnector.setInvisible(true);
			net.addArc(artificialSource, silentConnector);
			for(Place p : net.getPlaces())
				if(p.getLabel().equals("source 1")){
					net.addArc(silentConnector, p);
					break;
				}
			source = artificialSource;
		}
		if (net.getOutEdges(source).size() != 1) {
			throw new IllegalArgumentException("Invalid pattern: There is more than 1 start transition!");
		} else {
			return (Transition) net.getOutEdges(source).iterator().next().getTarget();
		}
	}
	
	public static Map<String, AbstractionPattern> buildPatterns(Map<DataPetriNet, String> patternToName) {
		Map<String, AbstractionPattern> abstractionPatterns = new HashMap<>();
		for (Entry<DataPetriNet, String> entry : patternToName.entrySet()) {
			AbstractionPattern pattern = new AbstractionPattern(entry.getValue(), entry.getKey());
			if(pattern.getDPN()==null)
				continue;
			pattern.setStartTransition(findStartTransition(pattern.getDPN()));
			pattern.setCompleteTransition(findCompleteTransition(pattern.getDPN()));
			for (Transition t : entry.getKey().getTransitions()) {
				pattern.getTransitions().add(t);
			}
			abstractionPatterns.put(pattern.getPatternName(), pattern);
		}
		return abstractionPatterns;
	}
	
	private static Transition findCompleteTransition(PetrinetGraph net) {
		Place sink = findSinkByStructure(net);
		if (net.getInEdges(sink).size() != 1) {
			throw new IllegalArgumentException("Invalid pattern: There is more than 1 complete transition!");
		} else {
			return (Transition) net.getInEdges(sink).iterator().next().getSource();
		}
	}
	
	private static Place findSource(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking initialMarking = ((DataPetriNetsWithMarkings) net).getInitialMarking();
			if (initialMarking != null) {
				if (initialMarking.size() == 1) {
					return initialMarking.iterator().next();
				} else {
					return null; // more than one place marked TODO
				}
			}
		}
		// Marking did not work try structure
		return findSourceByStructure(net);
	}
	
	public static DataPetriNet[] transformNets(PetrinetGraph[] abstractionNets) {
		DataPetriNet[] transformedNets = new DataPetriNet[abstractionNets.length];
		for (int i = 0; i < abstractionNets.length; i++) {
			transformedNets[i] = transformNet(abstractionNets[i]);
		}
		return transformedNets;
	}
	
	public static DataPetriNet transformNet(PetrinetGraph graph) {
 		PetriNetWithDataFactory factory = new PetriNetWithDataFactory(graph, graph.getLabel());
		PetriNetWithData transformedNet = factory.getRetValue();

		if (graph instanceof DataPetriNetsWithMarkings) {
			DataPetriNetsWithMarkings dpn = (DataPetriNetsWithMarkings) graph;
			if (dpn.getInitialMarking() != null) {
				transformedNet.setInitialMarking(factory.convertMarking(dpn.getInitialMarking()));
			}
			if (dpn.getFinalMarkings() != null) {
				Marking[] finalMarkings = dpn.getFinalMarkings();
				if (finalMarkings != null && finalMarkings.length > 0 && finalMarkings[0] != null) {
					transformedNet.setFinalMarkings(new Marking[] { factory.convertMarking(finalMarkings[0]) });
				}
			}
		}

		Place source = findSource(transformedNet);
		if (source == null) {
			Place artificialSource = graph.addPlace("artificialSource");
			Transition silentConnector = graph.addTransition("artificialSilentConnector");
			silentConnector.setInvisible(true);
			graph.addArc(artificialSource, silentConnector);
			for(Place p : graph.getPlaces())
				if(p.getLabel().equals("source 1")){
					graph.addArc(silentConnector, p);
					break;
				}
			source = artificialSource;
		}
		// choice after source place or loop back to source place requires invisible transition
		if (transformedNet.getOutEdges(source).size() > 1 || transformedNet.getInEdges(source).size() > 0) {
			Transition artificalStart = transformedNet.addTransition("artificalStart");
			artificalStart.setInvisible(true);
			Place artificalSource = transformedNet.addPlace("artificalSource");
			transformedNet.addArc(artificalSource, artificalStart);
			transformedNet.addArc(artificalStart, source);
		}

		Place sink = findSink(transformedNet);
		if (sink == null) {
			throw new IllegalArgumentException("Invalid pattern: Cannot find unique sink place!");
		}
		// choice before sink place or loop back from sink place requires invisible transition
		if (transformedNet.getInEdges(sink).size() > 1 || transformedNet.getOutEdges(sink).size() > 0) {
			Transition artificalEnd = transformedNet.addTransition("artificalEnd");
			artificalEnd.setInvisible(true);
			Place artificalSink = transformedNet.addPlace("artificalSink");
			transformedNet.addArc(sink, artificalEnd);
			transformedNet.addArc(artificalEnd, artificalSink);
		}

		if (graph instanceof DataPetriNet) {
			Map<DataElement, DataElement> oldVar2New = new HashMap<>();
			for (DataElement element : ((DataPetriNet) graph).getVariables()) {
				DataElement newVar = transformedNet.addVariable(element.getVarName(), element.getType(),
						element.getMinValue(), element.getMaxValue());
				oldVar2New.put(element, newVar);
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : graph.getEdges()) {
				if (edge instanceof VariableAccess) {
					if (edge.getSource() instanceof Transition) {
						DataElement oldVar = (DataElement) edge.getTarget();
						DataElement newVar = oldVar2New.get(oldVar);
						transformedNet.assignWriteOperation(factory.getTransMapping().get(edge.getSource()), newVar);
					}
					if (edge.getTarget() instanceof Transition) {
						DataElement oldVar = (DataElement) edge.getSource();
						DataElement newVar = oldVar2New.get(oldVar);
						transformedNet.assignReadOperation(factory.getTransMapping().get(edge.getSource()), newVar);
					}
				}
			}
			for (Transition t : graph.getTransitions()) {
				PNWDTransition transition = (PNWDTransition) factory.getTransMapping().get(t);
				if (((PNWDTransition) t).hasGuardExpression()) {
					try {
						transition.setGuard(transformedNet, ((PNWDTransition) t).getGuardExpression());
					} catch (NonExistingVariableException e) {
						throw new RuntimeException(e);
					}
				}
			}
		}

		return transformedNet;
	}
	
	private static Place findSink(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking[] finalMarkings = ((DataPetriNetsWithMarkings) net).getFinalMarkings();
			if (finalMarkings != null && finalMarkings.length > 0) {
				Marking finalMarking = ((DataPetriNetsWithMarkings) net).getFinalMarkings()[0];
				if (finalMarking != null) {
					if (finalMarking.size() == 1) {
						return finalMarking.iterator().next();
					} else {
						return null; // more than one place marked TODO
					}
				}
			}
		}
		// Marking did not work try structure
		return findSinkByStructure(net);
	}
	
	private static Place findSinkByStructure(PetrinetGraph net) {
		Place sink = null;
		for (Place p : net.getPlaces()) {
			if (net.getOutEdges(p).isEmpty()) {
				if (sink == null) {
					sink = p;
				} else {
					return null;
				}
			}
		}
		return sink;
	}
	
	public static Object[] abstractPatterns(XLog log, AbstractionModel model) throws ControlFlowAlignmentException,
	DataAlignmentException, PatternStructureException {

		BalancedProcessorConfiguration config = getConfiguration(model, log);
		
		// Usually the search space will be huge, so rather waste CPU cycles here instead of memory
		config.setActivateDataViewCache(false);
		config.setKeepDataFlowSearchSpace(false);
		
		XAlignedLog alignedLog = alignLogToAbstractionModel(new DummyPluginContext().getProgress(), config, log, model);
		LogAbstracterBasedOnAlignment abstractionProcessor = createAbstractionProcessor(model, config.getActivityMapping());
		
		DataPetriNetsWithMarkings dpnwm = model.getCombinedDPN();
		PetrinetWithMarkings pnwm = DataPetriNet.Factory.toPetrinetWithMarkings(dpnwm);
		Set<Marking> finalMarking = new HashSet<Marking>();
		for(Marking m : pnwm.getFinalMarkings())
			finalMarking.add(m);
		AcceptingPetriNet combinedModel = new AcceptingPetriNetImpl(pnwm.getNet(), pnwm.getInitialMarking(), finalMarking);
		try {
			abstractionProcessor.setErrorRateLimit(1);
			abstractionProcessor.setKeepUnmappedEvents(true);
			return abstractionProcessor.abstractBasedOnPatterns(alignedLog, combinedModel);
		} catch (PatternStructureException e) {
			throw e;
		}
	}
	
	public static Object[] abstractPatterns(final UIPluginContext context, XLog log, AbstractionModel model) throws UserCancelledException, ControlFlowAlignmentException,
			DataAlignmentException, PatternStructureException {

		BalancedProcessorConfiguration config = getConfiguration(model, log);

		// Usually the search space will be huge, so rather waste CPU cycles here instead of memory
		config.setActivateDataViewCache(false);
		config.setKeepDataFlowSearchSpace(false);

		config.setObserver(new DataConformancePlusObserverImpl(context) {

			public void startAlignment(int numExpectedResults) {
				super.startAlignment(numExpectedResults);
				context.getProgress().setMaximum(context.getProgress().getMaximum() + 1);
			}

		});

		XAlignedLog alignedLog = alignLogToAbstractionModel(new DummyPluginContext().getProgress(), config, log, model);
		LogAbstracterBasedOnAlignment abstractionProcessor = createAbstractionProcessor(model, config.getActivityMapping());

		System.err.println("all patterns used: "+!getExistsUnusedPatterns(alignedLog, model.getPatterns()));
		ProvidedObjectHelper.publish(context, "Abstraction: Alignment", alignedLog.getLog(), XLog.class, false);
		DataPetriNetsWithMarkings dpnwm = model.getCombinedDPN();
		ProvidedObjectHelper.publish(context, "Abstraction: Model1", dpnwm, DataPetriNetsWithMarkings.class, false);
		PetrinetWithMarkings pnwm = DataPetriNet.Factory.toPetrinetWithMarkings(dpnwm);
		Set<Marking> finalMarking = new HashSet<Marking>();
		for(Marking m : pnwm.getFinalMarkings())
			finalMarking.add(m);
		AcceptingPetriNet combinedModel = new AcceptingPetriNetImpl(pnwm.getNet(), pnwm.getInitialMarking(), finalMarking);
		ProvidedObjectHelper.publish(context, "Abstraction: Model2", combinedModel, AcceptingPetriNet.class, false);
		try {
			abstractionProcessor.setErrorRateLimit(1);
			abstractionProcessor.setKeepUnmappedEvents(true);
			return abstractionProcessor.abstractBasedOnPatterns(alignedLog, combinedModel);
		} catch (PatternStructureException e) {
			throw e;
		}
	}
	
	public static TransEvClassMapping transitionToEventClassMapperByLabel(final XLog log, final DataPetriNet net) {
		TransEvClassMapping map = new TransEvClassMapping(new XEventNameClassifier(), DUMMY);

		List<Transition> listTrans = new ArrayList<Transition>(net.getTransitions());
		Map<String, XEventClass> classes = new HashMap<String, XEventClass>();
		int index = 0;
		for (Transition transition : listTrans) {
			if (transition.isInvisible()) {
				map.put(transition, DUMMY);
			} else {
				XEventClass eClass = null;
				if(classes.containsKey(transition.getLabel()))
					eClass = classes.get(transition.getLabel());
				else{
					eClass = new XEventClass(transition.getLabel(), index);
					classes.put(transition.getLabel(), eClass);
					index++;
				}
				map.put(transition, eClass);
			}
		}
		return map;
	}
	
	// dummy event class (for unmapped transitions)
	public static final XEventClass DUMMY = new XEventClass("DUMMY", -1) {
		public boolean equals(Object o) {
			return this == o;
		}

		public int hashCode() {
			return System.identityHashCode(this);
		}		
	};
	
	private static BalancedProcessorConfiguration getConfiguration(AbstractionModel abstractionModel, XLog log) {

		DataPetriNetsWithMarkings net = abstractionModel.getCombinedDPN();

		BalancedProcessorConfiguration config = new BalancedProcessorConfiguration();
		
		TransEvClassMapping activityMapping = transitionToEventClassMapperByLabel(log, net);

		config.setActivityMapping(activityMapping);
		config.setInitialMarking(net.getInitialMarking());
		config.setFinalMarkings(net.getFinalMarkings());

		// populate event classes
		XEventClassifier classifier = config.getActivityMapping().getEventClassifier();
		XLogInfo info = log.getInfo(classifier);
		if (info == null) {
			info = XLogInfoFactory.createLogInfo(log, classifier);
		}
		XEventClasses eventClassesName = info.getEventClasses();
		List<XEventClass> evClasses = new ArrayList<>(eventClassesName.getClasses());
		evClasses.add(config.getActivityMapping().getDummyEventClass());
		ArrayList<Transition> transitionList = Lists.newArrayList(net.getTransitions());
		Collections.sort(evClasses, new Comparator<XEventClass>() {

			public int compare(XEventClass o1, XEventClass o2) {
				return o1.getId().compareTo(o2.getId());
			}
		});
		Collections.sort(transitionList, new Comparator<Transition>() {

			public int compare(Transition o1, Transition o2) {
				return o1.getLabel().compareTo(o2.getLabel());
			}
		});

		Map<XEventClass, Integer> mapEvClass2Cost = new HashMap<>();
		for (XEventClass evClass : evClasses) {
			if(evClass==DUMMY)
				mapEvClass2Cost.put(evClass, 0);
			else
				mapEvClass2Cost.put(evClass, 100);
		}
		config.setMapEvClass2Cost(mapEvClass2Cost);
		
		Map<Transition, Integer> costs = new HashMap<>();
		for (Transition trans : transitionList) {
			if(trans.isInvisible())
				costs.put(trans, 1);
			else
				costs.put(trans, 200000);
		}
		config.setMapTrans2Cost(costs);

		VariableMatchCosts variableCost = BalancedProcessorConfiguration.createDefaultVariableCost(net,
				Collections.<String>emptySet(), 0, 0);
		config.setVariableCost(variableCost);
		config.setVariableMapping(Collections.<String, String>emptyMap());

		return config;
	}
	
	private static BalancedProcessorConfiguration getConfiguration(UIPluginContext context,
			AbstractionModel abstractionModel, XLog log) {

		DataPetriNetsWithMarkings net = abstractionModel.getCombinedDPN();

		BalancedProcessorConfiguration config = new BalancedProcessorConfiguration();
		
		TransEvClassMapping activityMapping = transitionToEventClassMapperByLabel(log, net);

		config.setActivityMapping(activityMapping);
		config.setInitialMarking(net.getInitialMarking());
		config.setFinalMarkings(net.getFinalMarkings());

		// populate event classes
		XEventClassifier classifier = config.getActivityMapping().getEventClassifier();
		XLogInfo info = log.getInfo(classifier);
		if (info == null) {
			info = XLogInfoFactory.createLogInfo(log, classifier);
		}
		XEventClasses eventClassesName = info.getEventClasses();
		List<XEventClass> evClasses = new ArrayList<>(eventClassesName.getClasses());
		evClasses.add(config.getActivityMapping().getDummyEventClass());
		ArrayList<Transition> transitionList = Lists.newArrayList(net.getTransitions());
		Collections.sort(evClasses, new Comparator<XEventClass>() {

			public int compare(XEventClass o1, XEventClass o2) {
				return o1.getId().compareTo(o2.getId());
			}
		});
		Collections.sort(transitionList, new Comparator<Transition>() {

			public int compare(Transition o1, Transition o2) {
				return o1.getLabel().compareTo(o2.getLabel());
			}
		});

		Map<XEventClass, Integer> mapEvClass2Cost = new HashMap<>();
		for (XEventClass evClass : evClasses) {
			if(evClass==DUMMY)
				mapEvClass2Cost.put(evClass, 0);
			else
				mapEvClass2Cost.put(evClass, 100);
		}
		config.setMapEvClass2Cost(mapEvClass2Cost);
		
		Map<Transition, Integer> costs = new HashMap<>();
		for (Transition trans : transitionList) {
			if(trans.isInvisible())
				costs.put(trans, 1);
			else
				costs.put(trans, 100);
		}
		config.setMapTrans2Cost(costs);

		VariableMatchCosts variableCost = BalancedProcessorConfiguration.createDefaultVariableCost(net,
				Collections.<String>emptySet(), 0, 0);
		config.setVariableCost(variableCost);
		config.setVariableMapping(Collections.<String, String>emptyMap());

		return config;
	}


	public static boolean getExistsUnusedPatterns(XAlignedLog alignedLog, Collection<AbstractionPattern> patternMap) {
		Map<AbstractionPattern, Long> numMatches = new HashMap<>();
		for (AbstractionPattern pattern : patternMap) {
			long matchCount = 0;
			Transition startTransition = pattern.getStartTransition();
			String startTransitionId = startTransition.getLocalID().toString();
			for (XAlignment alignment : alignedLog) {
				for (XAlignmentMove move : alignment) {
					String activityId = move.getActivityId();
					if (startTransitionId.equals(activityId)) {
						matchCount++;
					}
				}
			}
			if(matchCount==0)
				return true;
			numMatches.put(pattern, matchCount);
		}
		return false;
	}
	
	public static XAlignedLog alignLogToAbstractionModel(final Progress progress,
			BalancedProcessorConfiguration alignmentConfig, XLog log, AbstractionModel model)
			throws ControlFlowAlignmentException, DataAlignmentException {
		BalancedDataXAlignmentPlugin alignmentPlugin = new BalancedDataXAlignmentPlugin();
		for (XEventClass eventClass : alignmentConfig.getActivityMapping().values()) {
			if (!alignmentConfig.getMapEvClass2Cost().containsKey(eventClass)) {
				alignmentConfig.getMapEvClass2Cost().put(eventClass, 1);
			}
		}
		XLog alignedLog = alignmentPlugin.alignLog(progress, model.getCombinedDPN(), log, alignmentConfig);
		return XAlignmentExtension.instance().extendLog(alignedLog);
	}
	
	public static LogAbstracterBasedOnAlignment createAbstractionProcessor(AbstractionModel abstractionModel,
			TransEvClassMapping transitionMapping) throws PatternStructureException {

		BiMap<Transition, String> startPoints = HashBiMap.create();
		BiMap<Transition, String> endPoints = HashBiMap.create();
		Map<Transition, Petrinet> originalNets = new HashMap<Transition, Petrinet>();
		//TODO support more than 'start' and 'complete'
		for (AbstractionPattern pattern : abstractionModel.getPatterns()) {
			// Translate all transitions
			pattern.setStartTransition((Transition) abstractionModel.getOld2NewMap().get(pattern.getStartTransition()));
			pattern.setCompleteTransition(
					(Transition) abstractionModel.getOld2NewMap().get(pattern.getCompleteTransition()));
			ImmutableList<Transition> oldTransitions = ImmutableList.copyOf(pattern.getTransitions());
			pattern.getTransitions().clear();
			for (Transition t : oldTransitions) {
				pattern.getTransitions().add((Transition) abstractionModel.getOld2NewMap().get(t));
			}
			// Configure abstraction with life-cycle transitions
			startPoints.put(pattern.getStartTransition(), pattern.getPatternName());
			endPoints.put(pattern.getCompleteTransition(), pattern.getPatternName());
			originalNets.put(pattern.getStartTransition(), DataPetriNet.Factory.toPetrinet(pattern.getDPN()));
		}

		return new LogAbstracterBasedOnAlignment(startPoints, endPoints, transitionMapping, originalNets);
	}
}
