package org.processmining.localprocessmodelconformance.util;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;

@Plugin(name = "Merge Identical Trace Variants", parameterLabels = { "Log" }, returnLabels = { "Log with Identical Traces Merged" }, returnTypes = { XLog.class }, userAccessible = true, categories = { PluginCategory.Filtering })
public class FromTracesToTraceVariants {
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl", pack = "LocalProcessModelConformance")
	@PluginVariant(variantLabel = "Merge Identical Trace Variants", requiredParameterLabels = { 0 })
	public XLog run(PluginContext context, XLog log) {
		XLog clone = (XLog) log.clone();
		return runHeadless(clone);
	}
	
	private XLog runHeadless(XLog log){
		Set<List<String>> traceVariantsSeen = new HashSet<List<String>>();
		Set<XTrace> tracesToRemove = new HashSet<XTrace>();
		for(XTrace trace : log){
			List<String> traceVariant = getActivityRepresentation(trace);
			if(traceVariantsSeen.contains(traceVariant))
				tracesToRemove.add(trace);
			
			traceVariantsSeen.add(traceVariant);
		}
		log.removeAll(tracesToRemove);
		return log;
	}
	
	public static String getActivity(XEvent event){
		return event.getAttributes().get("concept:name").toString();
	}
	
	private List<String> getActivityRepresentation(XTrace trace) {
		List<String> result = new ArrayList<String>();
		for(XEvent e : trace)
			result.add(getActivity(e));
		return result;
	}
}